#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>

using namespace std;

#define DEFAULT_FREQ 440
#define DEFAULT_VOL  100

static int Stereo;
static int Signed;
static int Bytes;
static int Rate;

enum {Sine,Square,Triangle} Form;

static class Generator *Generators = NULL;

class Generator
{
public:
    class Generator *Next;
public:
    double Tick;
    double DeltaTick;
    double Multi;
    double Volume;
    int Freq;
    
    Generator(Generator *anext, int afreq, int avol) :
        Next(anext), Volume(avol / 100.0), Freq(afreq)
    {
        /* May not setup more here, because the *
         * sampling rate and waveform are not   *
         * necessarily yet configured.          */
    }
    
    void SetupAll(void);    
    void DumpAll(const int _dummy=1);    
    float MixAll(int *_dummy = NULL);
};

void Generator::SetupAll(void)
{
    Multi     = (Form==Sine) ? M_PI+M_PI : 1.0;
    DeltaTick = Multi * (float)Freq / (float)Rate;
    
    Tick      = 0.0;
    
    if(Next)Next->SetupAll( );
}

void Generator::DumpAll(const int _dummy)
{
    if(Next)Next->DumpAll(0);    
    fprintf(stderr, "%d:%d%c", Freq, (int)(0.5+Volume*100), _dummy?'\n':' ');
}
    
float Generator::MixAll(int *_dummy)
{
    register double value = 0.0;
    double temp;
    
    switch(Form)
    {
        case Sine:
            value = sin(Tick);
            break;
        case Square:
            value = modf(Tick, &temp) < 0.5 ? -1 : 1;
            break;
        case Triangle:
            temp = modf(Tick, &temp) * 2.0;
            value = (temp<1.0 ? temp : 2.0 - temp) - 1.0;
            break;
    }
    
    value *= Volume;
    
    Tick += DeltaTick;
    
    if(!_dummy)
    {
        int Count=1;
        
        if(Next)value += Next->MixAll(&Count);
        
        value /= Count;
    }
    else
    {
        (*_dummy)++;
        if(Next)
            value += Next->MixAll(_dummy);
    }
    return value;
}

static void PrintVersionInfo(void)
{
    printf("%s",
        "Wave generator (wg) version " VERSION "\n"
        "Copyright (C) 1992,1999 Bisqwit (http://iki.fi/bisqwit/)\n"
    );
}

#include "argh.hh"

int main(int argc, const char *const *argv)
{
    ParamHandler Argh;
    Argh.AddLong("rate",     'r').SetInt(18,0x7FFFFFFF).SetDesc("Specify sampling rate. Default: -r44100.", "<num>");
    Argh.AddLong("bytes",    'b').SetInt(1,2).SetDesc("Specify byteness... 1=8-bit, 2=16-bit. Default: -b2.", "<num>");
    Argh.AddLong("mono",     'm').SetBool().SetDesc("Specify mono. This is the default.");
    Argh.AddLong("stereo",   's').SetBool().SetDesc("Specify stereo. Default: -m");
    Argh.AddLong("unsigned", 'u').SetBool().SetDesc("Specify unsigned samples. This is the default for -b1.");
    Argh.AddLong("signed",   'i').SetBool().SetDesc("Specify signed samples. This is the default for -b2.");
    Argh.AddLong("sine",     'S').SetBool().SetDesc("Select sine waveform. This is the default.");
    Argh.AddLong("square",   'Q').SetBool().SetDesc("Select square waveform. Default: -S.");
    Argh.AddLong("triangle", 'T').SetBool().SetDesc("Select triangle waveform. Default: -S.");
    Argh.AddLong("help",     'h').SetBool().SetDesc("Help");
    Argh.AddLong("version",  'V').SetBool().SetDesc("Version information");
    
    Stereo = Signed = Bytes = Rate = -1;
    Form = Sine;
    
    bool Heelp=false;
    Argh.StartParse(argc, argv);
    for(int c; (c=Argh.GetParam()) != -1; )
    	switch(c)
    	{
    		case 'V':
    			PrintVersionInfo();
    			exit(0);
    		case 'r': Rate = Argh.GetInt(); break;
    		case 'b': Bytes = Argh.GetInt(); break;
    		case 'm': Stereo = !Argh.GetBool(); break;
    		case 's': Stereo = Argh.GetBool(); break;
    		case 'u': Signed = !Argh.GetBool(); break;
    		case 'i': Signed = Argh.GetBool(); break;
    		case 'S': Form = Sine; break;
    		case 'Q': Form = Square; break;
    		case 'T': Form = Triangle; break;
    		case 'h': Heelp = true;
    		default:
    		{
    			string P = Argh.GetString();
    			const char *s = P.c_str();
                int Vol=DEFAULT_VOL, Freq;
                Freq = strtol(s, const_cast<char **> (&s), 10);
                if(*s==':')
                {
                    ++s;
                    Vol = strtol(s, const_cast<char **> (&s), 10);
                }
                Generators = new Generator(Generators, Freq, Vol);
    		}
    	}
    if(!Argh.ok())return -1;
    if(Heelp)
    {
    	PrintVersionInfo();
    	printf(
    	    "\nThis program works like a function generator.\n"
    	    "\nUsage: %s [<options> [<...>]] <frequency>[:<volume>] [<...>]\n\n", argv[0]);
    	Argh.ListOptions();
    	return 0;
    }
    if(Rate   < 0)Rate = 44100;
    if(Bytes  < 0)Bytes = 2;
    if(Stereo < 0)Stereo = 0;
    if(Signed < 0)Signed = (Bytes==2);
    if(!Generators)Generators = new Generator(Generators, DEFAULT_FREQ, DEFAULT_VOL);
    
    fprintf(stderr,
        "%s setup:   %d Hz %d-bit %s (%s) (%s wave)\n"
        "components: ",
        argv[0],
        Rate, Bytes*8, Stereo?"stereo":"mono",
        Signed?"signed":"unsigned",
        Form==Sine?"sine":
            Form==Square?"square":
                Form==Triangle?"triangle":
                    "random"
    );
    
    Generators->DumpAll();
    Generators->SetupAll();
    
    fflush(stderr);
        
    for(long Limit = 1 << (8*Bytes-1); ;)
    {
        int a, b;
        long Value, c;

        Value = (long)(Generators->MixAll() * (Limit-1));

        if(Value >= Limit)Value = Limit-1;
        if(Value < -Limit)Value = -Limit;
        
        if(!Signed)Value += Limit;

        for(a=Stereo; a>=0; a--)
            for(c=Value, b=Bytes; --b>=0; )
            {
                putchar(c & 255);
                c >>= 8;
            }
    }
    
    /* never reached, damn. */
    return 0;
}
